/***************************************************************************
 mapcat_main.cc  -  description
 -------------------
 begin                : Fri Dec  6 15:19:06 PST 2002
 copyright            : (C) 2002 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <list>
#include <sstream>
#include "functions.h"
#include "HSP.h"
#include "HSPMap.h"

#ifdef HAVE_PTHREAD
#include <math.h>
#include <pthread.h>
#include <stdexcept>
#include "ThreadPool.h"
#endif

using namespace std;

//#define DEBUG

//*******************************************
bool isAlu(const string &repName) {
   if (strncasecmp(repName.c_str(), "ALU", min((size_t) 3, repName.size())) == 0) {
      return true;
   }
   return false;
}

bool isSimple(const string& repName) {
   if (repName[repName.size() - 1] == 'n' && repName[repName.size() - 2] == ')') {
      return true;
   }
   return false;
}

template <class BidirectionalIterator>
BidirectionalIterator arrange(BidirectionalIterator beg, BidirectionalIterator end, BidirectionalIterator ptr) {
   using std::swap;
   bool done = false;
   for(BidirectionalIterator it(ptr); ptr != beg && *ptr < *--it; ptr = it) {
      swap(*ptr, *it);
      done = true;
   }
   if (!done) {
      for (BidirectionalIterator it(ptr); ptr != end && ++it != end && *ptr > *it; ptr = it) {
         swap(*ptr, *it);
      }
   }
   return ptr;
}

/**
 * Deletes either flanking or internal portions of the HSP pointed by lptr1,
 * that overlap with HSP pointed by lptr2.
 * Returns pointer which is at the same distance from the beginning as the
 * original lptr1 value.
 */
HSPMap::hsplist_t::iterator trim(HSPMap::hsplist_t& lst, HSPMap::hsplist_t::iterator lptr1, HSPMap::hsplist_t::iterator lptr2) {
   try {
      *lptr1 -= *lptr2;
#ifdef DEBUG
         cerr << *lptr1 << endl;
#endif
      arrange(lst.begin(), lst.end(), lptr1);
   } catch (logic_error err) {
      pair<HSP, HSP> flanks = (*lptr1).subsplit(*lptr2);
      if (flanks.second.getTop().getLength() > 0 && flanks.second.getBot().getLength() > 0) {
#ifdef DEBUG
         cerr << "Right:" << endl << flanks.second << endl;
#endif
         *lptr1 = flanks.second;
         arrange(lst.begin(), lst.end(), lptr1);
      }
      else {
         lptr1 = lst.erase(lptr1);
      }
      if (flanks.first.getTop().getLength() > 0 && flanks.first.getBot().getLength() > 0) {
#ifdef DEBUG
         cerr << "Left:" << endl << flanks.first << endl;
#endif
         lptr1 = arrange(lst.begin(), lst.end(), lst.insert(lptr1, flanks.first));
      }
   }
   return lptr1;
}

void defragment(HSPMap &rmap, float ovlpRate, bool doCombine, int minLen, float minSim, float minScore);
void defragment(HSPMap::hsplist_t &lst, float ovlpRate, bool doCombine, int minLen, float minSim, float minScore);
#ifdef HAVE_PTHREAD

struct defragment_params {
   HSPMap &rmap;
   HSPMap::iterator rmptr;
   float ovlpRate;
   bool doCombine;
   int minLen;
   float minSim;
   float minScore;
   pthread_mutex_t write_mutex;
   defragment_params(HSPMap &rmap, float ovlpRate, bool doCombine, int minLen, float minSim, float minScore) :
   rmap(rmap), ovlpRate(ovlpRate), doCombine(doCombine), minLen(minLen), minSim(minSim), minScore(minScore), rmptr(rmap.begin()) {
      if (pthread_mutex_init(&write_mutex, NULL) != 0) {
         throw(runtime_error("Failed to initialize write_mutex"));
      }
   }
   ~defragment_params() {
      pthread_mutex_destroy(&write_mutex);
   }
};
void* defragment_thread(void* args) {
   defragment_params* params = (defragment_params*)args;
   HSPMap::iterator mptr;
   while(pthread_mutex_lock(&params->write_mutex) == 0 && params->rmptr != params->rmap.end()) {
      mptr = params->rmptr;
      ++(params->rmptr);
      pthread_mutex_unlock(&params->write_mutex);
      defragment((*mptr).second, params->ovlpRate, params->doCombine, params->minLen, params->minSim, params->minScore);
   }
   pthread_mutex_unlock(&params->write_mutex);
}
#endif
//*******************************************
void defragment(HSPMap &rmap, float ovlpRate, bool doCombine, int minLen, float minSim, float minScore) {
#ifdef HAVE_PTHREAD
   if(rmap.size() > 1) {
      ThreadPool tp(min(rmap.size(), ThreadPool::getCoreCount()));
      defragment_params params(rmap, ovlpRate, doCombine, minLen, minSim, minScore);
      tp.initWorkers(defragment_thread, &params);
      tp.joinWorkers();
   }
   else
#endif
   {
      HSPMap::iterator rmptr;
      for (rmptr = rmap.begin(); rmptr != rmap.end(); rmptr++) {
         defragment((*rmptr).second, ovlpRate, doCombine, minLen, minSim, minScore);
      }
   }
}

void defragment(HSPMap::hsplist_t &lst, float ovlpRate, bool doCombine, int minLen, float minSim, float minScore) {
   HSPMap::hsplist_t::iterator lptr1, lptr2;
   //====================================================================
   // First loop
   //====================================================================
#ifdef DEBUG
   cerr << "Loop 1\n";
#endif
   for (lptr1 = lst.begin(); lptr1 != lst.end(); lptr1++) {
      for (lptr2 = lptr1, lptr2++; lptr2 != lst.end();) {
         if (lptr1 == lptr2) {
            throw(logic_error("Comparing HSP to itself."));
         }
         //Appending poly-A tail, if any, to the end of ALU
         if (doCombine) {
            if ((*lptr1).isDirect() && (*lptr2).isDirect()) {
               if (dist((*lptr1).getTop(), (*lptr2).getTop(), false) < 40) {
                  if (isAlu((*lptr1).getBot().getName()) && (*lptr2).getBot().getName() == "(A)n") {
                     *lptr1 += *lptr2;
                     lptr2 = lst.erase(lptr2);
                     continue;
                  }
               }
            }
            if (!(*lptr1).isDirect() && !(*lptr2).isDirect()) {
               if (dist((*lptr1).getTop(), (*lptr2).getTop(), false) < 40) {
                  if (isAlu((*lptr2).getBot().getName()) && (*lptr1).getBot().getName() == "(A)n") {
                     *lptr2 += *lptr1;
                     lst.erase(lptr1);
                     lptr2 = lptr1 = arrange(lst.begin(), lst.end(), lptr2);
                     ++lptr2;
                     continue;
                  }
               }
            }
         }
         //Joining together 2 simple repeats, if close to each other
         // dist() automatically checks for names' equality and throws
         // logic_error if names are different
         try {
            if ((*lptr1).isDirect() && (*lptr2).isDirect()) {
               if (dist((*lptr1).getTop(), (*lptr2).getTop()) <= 6) {
                  if (isSimple((*lptr1).getBot().getName()) && isSimple((*lptr2).getBot().getName())) {
                     *lptr1 += *lptr2;
                     lptr2 = lst.erase(lptr2);
                     continue;
                  }
               }
            }
         } catch (logic_error err) {
         }
         //Joining 2 fragments of the same repeat, if close enough
         if (areJoinable(*lptr1, *lptr2)) {
#ifdef DEBUG
            cerr << *lptr1 << endl;
            cerr << *lptr2 << endl;
            fprintf(stderr, "=>Close to\n");
#endif
            *lptr1 += *lptr2;
            lptr2 = lst.erase(lptr2);
#ifdef DEBUG
            cerr << *lptr1 << endl << endl;
#endif
            continue;
         }

         if (dist((*lptr1).getTop(), (*lptr2).getTop(), false) > MAXDIST) {
            break;
         }
         lptr2++;
      }
   }
   //====================================================================
   // Second loop
   //====================================================================
   for (lptr1 = lst.begin(); lptr1 != lst.end(); lptr1++) {
      //Removing overlaping of 2 fragments by partially erasing one of them
      for (lptr2 = lptr1, lptr2++; lptr2 != lst.end();) {
         if (lptr1 == lptr2) {
            throw(logic_error("Comparing HSP to itself."));
         }
         //Removing overlaping of 2 fragments by erasing one of them
         if (overlap((*lptr1).getTop(), (*lptr2).getTop(), 0.0, false)) {
#ifdef DEBUG
               cerr << endl << "Loop 2:" << endl;
               cerr << *lptr1 << endl;
               cerr << *lptr2 << endl;
#endif
            // If fragments overlap by almost full length specified by minimum
            // value of ovlpRate which must be close to 1.0
            if (overlap((*lptr1).getTop(), (*lptr2).getTop(), ovlpRate, false)) {
#ifdef DEBUG
               cerr << (overlap((*lptr1).getTop(), (*lptr2).getTop(), 0, false) ? "overlap" : "don't overlap") << ", " << (*lptr1).getTop().getBeg() << ", dist = " << dist((*lptr1).getTop(), (*lptr2).getTop()) << "\n";
               fprintf(stderr, "=>Overlaps Over %f\n", ovlpRate);
#endif

               if ((*lptr1).getScore() < (*lptr2).getScore() || ((*lptr1).getScore() == (*lptr2).getScore() && (*lptr1).getSim() <= (*lptr2).getSim())) {
                  if ((*lptr1).getTop().getLength() <= (*lptr2).getTop().getLength()) {
#ifdef DEBUG
                     cerr << "(2)Erase *lptr1" << endl;
#endif
                     lptr1 = lst.erase(lptr1);
                     lptr2 = lptr1;
                     lptr2++;
                  }
                  else {
#ifdef DEBUG
                     cerr << "(2)Trimming *lptr1" << endl;
#endif
                     lptr2 = lptr1 = trim(lst, lptr1, lptr2);
                     ++lptr2;
                  }
               } else {
                  if ((*lptr1).getTop().getLength() >= (*lptr2).getTop().getLength()) {
#ifdef DEBUG
                     cerr << "(2)Erase *lptr2" << endl;
#endif
                     lptr2 = lst.erase(lptr2);
                  }
                  else {
#ifdef DEBUG
                     cerr << "(2)Trimming *lptr2" << endl;
#endif
                     lptr2 = trim(lst, lptr2, lptr1);
                  }
               }
               continue;
            }
         } else {
            break;
         }
         lptr2++;
      }
   }

   //====================================================================
   // Third loop
   //====================================================================
   for (lptr1 = lst.begin(); lptr1 != lst.end();) {
      //Removing overlaping of 2 fragments by partially erasing one of them
      for (lptr2 = lptr1, ++lptr2; lptr2 != lst.end();) {
         if (lptr1 == lptr2) {
            throw(logic_error("Comparing HSP to itself."));
         }
#ifdef DEBUG
         cerr << endl << "Loop 3:" << endl;
         cerr << *lptr1 << endl;
         cerr << *lptr2 << endl;
         cerr << (*lptr1).getTop().getBeg() << ", dist = " << dist((*lptr1).getTop(), (*lptr2).getTop()) << "\n";
#endif

         if (overlap((*lptr1).getTop(), (*lptr2).getTop())) {
#ifdef DEBUG
            fprintf(stderr, "=>Overlaps\n");
#endif
            if (contain((*lptr1).getTop(), (*lptr2).getTop())) {
#ifdef DEBUG
               cerr << (*lptr1) << endl;
#endif
               lptr2 = lst.erase(lptr2);
               continue;
            }
            if (contain((*lptr2).getTop(), (*lptr1).getTop())) {
#ifdef DEBUG
               cerr << (*lptr2) << endl;
#endif
               lptr2 = lptr1 = lst.erase(lptr1);
               ++lptr2;
               continue;
            }
            if ((*lptr1).getScore() < (*lptr2).getScore() || (*lptr1).getScore() == (*lptr2).getScore() && (*lptr1).getTop().getLength() >= (*lptr2).getTop().getLength()) {
#ifdef DEBUG
               cerr << "(3)Trimming *lptr1" << endl;
#endif
               lptr2 = lptr1 = trim(lst, lptr1, lptr2);
               ++lptr2;
               continue;
            } else {
#ifdef DEBUG
               cerr << "(3)Trimming *lptr2" << endl;
#endif
               lptr2 = trim(lst, lptr2, lptr1);
               continue;
            }
         } else {
            break;
         }
      }
      int iLT = minLen;
      // Removing short fragments, that are most likely to be garbage
      if (isSimple((*lptr1).getBot().getName())) {
         iLT = 10;
      }
      if ((*lptr1).getTop().getLength() * (*lptr1).getSim() < iLT || (*lptr1).getSim() < minSim || (*lptr1).getScore() < minScore) {
#ifdef DEBUG
         cerr << (*lptr1).getTop().getLength() << " * " << (*lptr1).getSim() << " = " << (*lptr1).getTop().getLength() * (*lptr1).getSim() << "\n";
         cerr << "Erasing as insignificant: " << (*lptr1) << "\n";
#endif
         lptr1 = lst.erase(lptr1);
      }
      else {
         ++lptr1;
      }
   }
}

#ifdef DEBUG
#include <sstream>
void testHSP() {
   string input = "chr221        15351090 15352684 MST_I             99 1636    c   0.7377  0.7400  5264 \n\
chr221        15351944 15352841 MST_I              1 873     c   0.7150  0.7200  2391 \n\
chr222        20567235 20568462 SVA              424 1640    c   0.9165 0.92      9573    2.47   74.21 \n\
chr222        20567972 20569139 SVA                1 1127    c   0.9367 0.94      7776    2.35   68.72 \n\
ref|NT_022517.18|:1-66110270    4199 4368    L1ME4A           233 407     d   0.6744  0.6700  458 \n\
ref|NT_022517.18|:1-66110270    4440 4693    L1MA1            758 1022    d   0.6988  0.7000  491 \n\
ref|NT_022517.18|:1-66110270    4443 4693    L1MA3            753 1024    d   0.7619  0.7600  979 \n\
ref|NT_022517.18|:1-66110270    4444 4696    L1MA4            745 1023    d   0.8366  0.8400  1360 \n\
ref|NT_022517.18|:1-66110270    4444 4711    L1MA5            745 1047    d   0.8139  0.8100  1278 \n\
ref|NT_022517.18|:1-66110270    4717 4970    L1ME_ORF2       1992 2257    d   0.6628  0.6700  469 \n\
";
   istringstream iss(input);
   HSPMap rmap;
   loadMap(iss, rmap, true);
   printMap(cout, rmap);
   defragment(rmap, 0.999, true, 20, 0.6, 0);
   cout << endl;
   printMap(cout, rmap);
}
#endif

//*******************************************
void usage(char *exec) {
   if (!exec) {
      return;
   }
   if (strrchr(exec, '/')) {
      exec = strrchr(exec, '/') + 1;
   }
   cerr << "NAME\n\t" << exec;
#ifdef VERSION

   cerr << " V" << VERSION;
#endif

   cerr << " - removes overlapings and joins close fragments on the map\n";
   cerr << "\nSYNOPSIS\n\t" << exec << " [in_map(-)][out_map][-nc][--help][--version]\n";
   cerr << "\t[-noself] [-noselfonly] [-bylength] [-minlen #]\n";
   cerr << "\nDESCRIPTION\n\tRemoves overlapings and joins fragments that logically\n";
   cerr << "\tcould be fragments of the same element on the map\n";
   cerr << "\nFLAGS\n";
   cerr << "\t-minlen #\n\t   Minimum allowed length of scoring pairs reported in map.\n";
   cerr << "\n\t-minsim #\n\t   Minimum allowed similarity of scoring pairs reported in map.\n";
   cerr << "\n\t-minscore #\n\t   Minimum allowed score of scoring pairs reported in map.\n";
   cerr << "\n\t-qut n|p\n\t   Query sequence is nucleotide/protein. Default is nucleotide.\n";
   cerr << "\n\t-dbt n|p\n\t   Database sequence is nucleotide/protein. Default is nucleotide.\n";
   cerr << "\nFLAGS\n";
   cerr << "\t-nc\n\t   do not combine.\n";
   cerr << "\t   The program will not append poly-A tail to ALU repeats.\n";
   cerr << "\n\t-noself\n\t   removes self-hits from the map before processing.\n";
   cerr << "\n\t-noselfonly\n\t   only removes self-hits from the map and exits.\n\n";
#ifdef PACKAGE_BUGREPORT

   cerr << "\nBUG REPORT\n\t" << PACKAGE_BUGREPORT << "\n";
#endif
#ifdef COPYRIGHT

   cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}

//*******************************************
int main(int argc, char** argv) {
   //testHSP(); exit(0);
   istream *is;
   ostream *os;
   char** args = argv;
   bool doCombine = true;
   bool noSelf = false;
   bool noSelfOnly = false;
   bool isQUp = false;
   bool isDBp = false;
   int minLen = 21;
   float minSim = 0.2;
   float minScore = 0.0;
   ARGMAP amap;
   getflags(argv, amap, 5, "-noself", "-noselfonly", "-nc", "--help", "--version");
   getoptions(argv, amap, 5, "-qut", "-dbt", "-minlen", "-minsim", "minscore");
   if (amap.find("-qut") != amap.end()) {
      if (amap["-qut"] == "p" || amap["-qut"] == "P") {
         isQUp = true;
      }
   }
   if (amap.find("-dbt") != amap.end()) {
      if (amap["-dbt"] == "p" || amap["-dbt"] == "P") {
         isDBp = true;
      }
   }
   if (amap.find("-minlen") != amap.end()) {
      if (sscanf(amap["-minlen"].c_str(), "%d", &minLen) < 0 || minLen < 0) {
         fprintf(stderr, "Bad value of -minlen option.\n");
         exit(5);
      }
   }
   if (amap.find("-minsim") != amap.end()) {
      if (sscanf(amap["-minsim"].c_str(), "%f", &minSim) < 0 || minSim < 0) {
         fprintf(stderr, "Bad value of -minsim option.\n");
         exit(6);
      }
   }
   if (amap.find("-minscore") != amap.end()) {
      if (sscanf(amap["-minscore"].c_str(), "%f", &minScore) < 0 || minScore < 0) {
         fprintf(stderr, "Bad value of -minscore option.\n");
         exit(7);
      }
   }
   if (amap.find("-nc") != amap.end()) {
      doCombine = false;
   }
   if (amap.find("-noself") != amap.end()) {
      noSelf = true;
   }
   if (amap.find("-noselfonly") != amap.end()) {
      noSelfOnly = true;
   }
   if (amap.find("--help") != amap.end() || amap.find("--version") != amap.end()) {
      usage(argv[0]);
      exit(0);
   }
   if ((args = assigninput(is, argv)) == 0) {
      fprintf(stderr, "Failed to asign input.\n");
      usage(argv[0]);
      exit(10);
   }
   HSPMap hspmap(isDBp, isQUp);
   if(hspmap.read(*is, noSelf || noSelfOnly).size() == 0) {
      fprintf(stderr, "%s: Read empty map\n", argv[0]);
   }
   if(is != &cin) {
      delete is;
   }
   if ((args = assignoutput(os, args)) == 0) {
      fprintf(stderr, "Failed to asign output.\n");
      usage(argv[0]);
      exit(30);
   }
   //=============================================
   int res = 0;
   if (!noSelfOnly) {
      defragment(hspmap, 0.999, doCombine, minLen, minSim, minScore);
   }
   hspmap.write(*os);
   if(os != &cout) {
      delete os;
   }
   //=============================================
   return 0;
}
