#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "constants.h"
#include "functions.h"

#define MAXLEN 4000
#define FRAGMENT "FRAGMENT"

/** ======================================================== */
void usage(char* exec)
{
    if(!exec)
    {
        return;
    }
    if(strrchr(exec, '/'))
    {
        exec = strrchr(exec, '/') + 1;
    }
    cerr << "NAME\n\t" << exec;
#ifdef VERSION

    cerr << " V" << VERSION;
#endif

    cerr << " - converts IG formated sequences into FASTA format\n";
    cerr << "\nSYNOPSIS\n";
    cerr << "\t" << exec << " [ig_input(-)] [fasta_output]\n";
    cerr << "\t[-n len]\n";
    cerr << "\nDESCRIPTION\n";
    cerr << "\tConverts IG formated sequences into FASTA format\n";
    cerr << "\nOPTIONS\n";
    cerr << "\t-n # - number of bases per data line in the output.\n";
    cerr << "\nFLAGS\n";
    cerr << "\t-ol - onle line per sequence. Each sequence body will be in one line.\n";
    cerr << "\t-uc - converts sequences to upper case.\n";
    cerr << "\t-unseq - eliminates N from the sequence, in order to clean unsequenced data.\n";
    cerr << "\nAUTHOR\n";
    cerr << "\tOleksiy Kohany\n";
#ifdef PACKAGE_BUGREPORT

    cerr << "\nBUG REPORT\n\t" << PACKAGE_BUGREPORT << "\n";
#endif
#ifdef COPYRIGHT

    cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}

int main(int argc, char **argv)
{
    char **args;
    FILE *FIG, *FFASTA;
    int lineLen = LINELENGTH;
    map<string, int> umap;
    bool noN = false;
    bool toUpper = false;
    bool oneLine = false;
    ARGMAP amap;
    ::getoptions(argv, amap, 1, "-n");
    ::getflags(argv, amap, 3, "-unseq", "-uc", "-ol");
    if(amap.find("-uc") != amap.end())
    {
        toUpper = true;
    }
    if(amap.find("-ol") != amap.end())
    {
        oneLine = true;
    }
    if(amap.find("-unseq") != amap.end())
    {
        noN = true;
    }
    if(amap.find("-n") != amap.end())
    {
        if(sscanf(amap["-n"].c_str(), "%d", &lineLen) < 1)
        {
            fprintf(stderr, "Bad argument '%s' to -n option\n", amap["-n"].c_str());
            usage(argv[0]);
            exit(10);
        }
        if(lineLen < 5)
        {
            lineLen = LINELENGTH;
        }
    }
    if((args = assigninput(FIG, argv)) == 0)
    {
        fprintf(stderr, "Failed to asign input.\n");
        usage(argv[0]);
        exit(10);
    }
    if((args = assignoutput(FFASTA, args)) == 0)
    {
        fprintf(stderr, "Failed to asign output.\n");
        usage(argv[0]);
        exit(20);
    }
    //===========================================
    char sLine[MAXLEN];
    char sName[MAXNAMESIZE];
    char sFrag[MAXLEN] = "";
    char *ptr;
    long iCnt = 0;
    for(ptr = fgets(sLine, MAXLEN, FIG); ptr != 0; ptr = fgets(sLine, MAXLEN, FIG))
    {
        if(strlen(ptr) > 1)
        {
            sFrag[0] = 0;
            if(*ptr != ';' && *ptr != '>')
            {
                fprintf(stderr, "File is not in IG or FASTA format.\n");
                exit(3);
            }
            break;
        }
    }
    if(ptr == 0)
    {
        exit(4);
    }
    do
    {
        switch(*ptr)
        {
        case ';':
            do
            {
                if((ptr = strstr(sLine, FRAGMENT)) != 0 && strstr(sLine, "->") != 0)
                {
                    int f, t;
                    if(sscanf(ptr, "%s %d -> %d", sFrag, &f, &t) == 3)
                    {
                        strcpy(sFrag, ptr);
                    }
                }
            }
            while((ptr = fgets(sLine, MAXLEN, FIG)) != 0 && sLine[0] == ';');
            if(ptr == 0)
            {
                exit(0);
            }
            sscanf(ptr, "%s", sName);
            if(iCnt > 0)
            {
                fprintf(FFASTA, "\n");
            }
            fprintf(FFASTA, ">%s", sName);
            if(sFrag[0] != 0)
            {
                fprintf(FFASTA, " %s", sFrag);
                sFrag[0] = '\0';
            }
            else
            {
                fprintf(FFASTA, "\n");
            }
            iCnt = 0;
            break;
        case '>':
            if(iCnt > 0)
            {
                fprintf(FFASTA, "\n");
                iCnt = 0;
            }
            fprintf(FFASTA, "%s", ptr);
            break;
        default:
            for(; *ptr != '\0'; ptr++)
            {
                if((isalpha(*ptr) || *ptr=='-') && *ptr != '1')
                {
                    if(noN && (*ptr == 'n' || *ptr == 'N'))
                    {
                        continue;
                    }
                    if(iCnt >= lineLen)
                    {
                        if(!oneLine)
                        {
                            fprintf(FFASTA, "\n");
                        }
                        iCnt = 0;
                    }
                    if(toUpper)
                    {
                        fprintf(FFASTA, "%c", toupper(*ptr));
                    }
                    else
                    {
                        fprintf(FFASTA, "%c", *ptr);
                    }
                    iCnt++;
                }
            }
        }
    }
    while((ptr = fgets(sLine, MAXLEN, FIG)) != 0);
    fprintf(FFASTA, "\n");
    //===========================================
    FIG == stdin || fclose(FIG);
    FFASTA == stdout || fclose(FFASTA);
}
