/*
 * Fragment.cpp
 *
 *  Created on: Feb 19, 2010
 *      Author: kohany
 */

#include "Fragment.h"
#include <stdlib.h>
#include <iomanip>
#include <stdexcept>

#ifdef HAVE_PTHREAD
   concurrent_set<string> Fragment::cset_;
#else
   set<string> Fragment::cset_;
#endif

Fragment::Fragment() :
   pName(NULL), beg(0), end(0), direct(true) {
}

Fragment::Fragment(const string& name, size_t beg, size_t end) :
   pName(&*(cset_.insert(name).first)),
   beg(beg),
   end(end) {
   validate();
}

Fragment::Fragment(const Fragment& other) :
   pName(&*(cset_.insert(other.getName()).first)),
   beg(other.getBeg()),
   end(other.getEnd()),
   direct(other.isDirect()) {
}

Fragment::~Fragment() {
}

Fragment& Fragment::operator =(const Fragment& other) {
   pName = &*(cset_.insert(other.getName()).first);
   beg = other.getBeg();
   end = other.getEnd();
   direct = other.isDirect();
   return *this;
}
void Fragment::assign(const string& name, size_t beg, size_t end) {
   pName = &*(cset_.insert(name).first);
   assign(beg, end);
}
void Fragment::assign(size_t beg, size_t end) {
   this->beg = beg;
   this->end = end;
   validate();
}

void Fragment::validate() {
   if (beg > end) {
      std::swap(beg, end);
      direct = false;
   }
   else {
      direct = true;
   }
}

void Fragment::reset() {
   pName = NULL;
   beg = end = 0;
   direct = true;
}

const string& Fragment::getName() const {
   if(pName) {
      return *pName;
   }
   throw logic_error("Request for uninitialized name.");
}

Fragment& Fragment::operator +=(const Fragment& other) {
   beg = min(getBeg(), other.getBeg());
   end = max(getEnd(), other.getEnd());
   return *this;
}
Fragment& Fragment::operator -=(const Fragment& other) {
   if(overlap(*this, other)) {
      if(contain(other, *this, false)) {
         reset();
      }
      else if(contain(*this, other, false) && getBeg() != other.getBeg() && getEnd() != other.getEnd()) {
         throw(logic_error("Subtracting internal region"));
      }
      else {
         if(getBeg() < other.getBeg() && getEnd() <= other.getEnd()) {
            end = other.getBeg();
         }
         if(getBeg() >= other.getBeg() && getEnd() > other.getEnd()) {
            beg = other.getEnd();
         }
      }
   }
   return *this;
}
Fragment& Fragment::operator +=(const size_t v) {
   beg += v;
   end += v;
   return *this;
}
Fragment& Fragment::operator -=(const size_t v) {
   beg -= v;
   end -= v;
   return *this;
}
Fragment& Fragment::operator *=(const double v) {
   beg *= v;
   end *= v;
   return *this;
}
Fragment& Fragment::operator /=(const double v) {
   beg /= v;
   end /= v;
   return *this;
}
void Fragment::resize(ssize_t left, ssize_t right) {
   if(left < 0) {
      if(right < 0) {
         shrink(-left, -right);
      }
      else {
         shrink(-left, 0);
         expand(0, right);
      }
   }
   else {
      if(right >= 0) {
         expand(left, right);
      }
      else {
         expand(left, 0);
         shrink(0, -right);
      }
   }
}
void Fragment::expand(size_t left, size_t right) {
   if(isDirect()) {
      beg = beg < left ? 0 : beg - left;
      end += right;
   }
   else {
      beg = beg < right ? 0 : beg - right;
      end += left;
   }
   if(beg > end) {
      end = beg;
   }
}
void Fragment::shrink(size_t left, size_t right) {
   if(isDirect()) {
      beg += left;
      end = end < right ? 0 : end - right;
   }
   else {
      beg += right;
      end = end < left ? 0 : end - left;
   }
   if(beg > end) {
      end = beg;
   }
}
pair<Fragment, Fragment> Fragment::subsplit(const Fragment& other) const {
   pair<Fragment, Fragment> flanks;
   using std::swap;
   if(contain(*this, other)) {
      flanks.first.assign(getName(), getBeg(), other.getBeg());
      flanks.second.assign(getName(), other.getEnd(), getEnd());
   }
   else {
      flanks.first = *this;
      flanks.first -= other;
      flanks.second.assign(getName(), 0, 0);
      if(getBeg() < flanks.first.getBeg()) {
         swap(flanks.first, flanks.second);
      }
      if(getEnd() > flanks.first.getEnd()) {
         swap(flanks.first, flanks.second);
      }
   }
   return flanks;
}
const bool Fragment::equals(const Fragment& other) const {
   if(getBeg() == other.getBeg() && getEnd() == other.getEnd() && getName() == other.getName()) {
      return true;
   }
   return false;
}
Fragment& Fragment::read(istream& is) {
   string name;
   size_t beg = 0;
   size_t end = 0;
   is >> name >> beg >> end;
   if(beg <= end) {
      assign(name, beg - 1, end);
   }
   else {
      assign(name, beg, end - 1);
   }
   return *this;
}
void Fragment::write(ostream& os) const {
   os << setw(12) << left << getName() << " "<< setw(7) << right << getBeg() + 1 << " " << setw(7) << left << getEnd() << " ";
}

//================================================================
istream& operator >> (istream& is, Fragment& f) {
   f.read(is);
   return is;
}
ostream& operator << (ostream& os, const Fragment& f) {
   f.write(os);
   return os;
}

const ssize_t dist(const Fragment& f1, const Fragment& f2, bool compareNames) {
   if(compareNames && f1.getName().compare(f2.getName())) {
      throw logic_error("Distance calculation on different sequences.");
   }
   return max(f1.getBeg(), f2.getBeg()) - min(f1.getEnd(), f2.getEnd());
}

const ssize_t range(const Fragment& f1, const Fragment& f2, bool compareNames) {
   if(compareNames && f1.getName().compare(f2.getName())) {
      throw logic_error("Range calculation on different sequences.");
   }
   return min(f1.getBeg(), f2.getBeg()) - max(f1.getEnd(), f2.getEnd());
}

const bool overlap(const Fragment& f1, const Fragment& f2, double threshold, bool compareNames) {
   try {
      ssize_t d = dist(f1, f2, compareNames);
      if(d < 0) {
         if(threshold > 0 && double(-d) / min(f1.getLength(), f2.getLength()) < threshold) {
            return false;
         }
         return true;
      }
   }
   catch(logic_error err) {
   }
   return false;
}
const bool contain(const Fragment& f1, const Fragment& f2, bool compareNames) {
   if(f1.getBeg() <= f2.getBeg() && f1.getEnd() >= f2.getEnd()) {
      if(compareNames && f1.getName().compare(f2.getName())) {
         return false;
      }
      return true;
   }
   return false;
}
const bool areJoinable(const Fragment& f1, const Fragment& f2, int maxDist, bool compareNames) {
   if(compareNames && f1.getName().compare(f2.getName()) || f1.isDirect() ^ f2.isDirect()) {
      return false;
   }
   ssize_t d = dist(f1, f2, false);
   if(d <= maxDist && abs(d) < f1.getLength() && abs(d) < f2.getLength()) {
      return true;
   }
   return false;
}
const bool areOrdered(const Fragment& f1, const Fragment& f2, bool compareNames) {
   if(compareNames && f1.getName().compare(f2.getName())) {
      return false;
   }
   if(f1.getRBeg() <= min(f1.getREnd(), f2.getRBeg()) && max(f1.getREnd(), f2.getRBeg()) <= f2.getREnd()) {
      return true;
   }
   if(f1.getRBeg() >= max(f1.getREnd(), f2.getRBeg()) && min(f1.getREnd(), f2.getRBeg()) >= f2.getREnd()) {
      return true;
   }
   return false;
}
